#include "OLED.h"

#define OLED_WIDTH 128  // OLEDの幅(ピクセル数)
#define OLED_HEIGHT 64  // OLEDの高さ(ピクセル数)
#define OLED_PAGES 8    // OLEDのページ数(ピクセル数÷８)

#define OLED_CHAR_W 16
#define OLED_CHAR_H 4

static uint8_t oledAddr = 0x3C;     // SSD1306のI2Cアドレス
static uint8_t oldCommand = 0x80;   // SSD1306の1Byteコマンド
static uint8_t oldCommands = 0x00;  // SSD1306の複数Byteコマンド
static uint8_t oldData = 0x40;      // SSD1306のデータバイト
static uint8_t DISP7SEG_SSD1306_INIT[] = { // SSD1306の初期化コマンド群
  0xC8,         // C0/C8:上下の描画方向
  0xA1,         // A0/A1:左右の描画方向
  0x22,0x00,0x07,         // 20/21/22/23:アドレッシングモード
  0xA8,0x3F,    // 画面の解像度
  0xD3,0x00,    // 縦方向のオフセット
  0xDA,0x12,    // 描画タイプ
  0x40,         // 縦方向の描画開始位置
  0xA6,         // 表示方法 - A7にすると画面反転
  0x8D, 0x14    // OLEDパネルの供給電圧と思われる 14hに設定しないとダメらしい
};

// 初期化
void OLED::init() {
  // KANJI ROM Init
  kanjiRom.init() ;
  
  // OLED設定コマンド送信
  Wire.begin();

  oledSendCommands(sizeof(DISP7SEG_SSD1306_INIT), DISP7SEG_SSD1306_INIT) ;  // 初期化
  oledSendCommand(0xAF) ; // 画面表示
  cls() ;
  dispSwitch(true) ;
}

// 画面消去
void OLED::cls() {
  // 表示位置初期化
  cursorX = 0 ;
  cursorY = 0 ;

  // スクロールバッファ消去
  memset(DisplayBuff,0,sizeof(DisplayBuff)) ;

  clsOLED() ;
}

void OLED::clsOLED() {
  // 画面消去
  oledSendCommand(0xAE) ; // 画面表示

  for (uint8_t m = 0; m < OLED_PAGES; m++) {
    oledSendCommand3(0xB0 | (m & 0x07) , 0x10 , 0x00) ;
    Wire.beginTransmission(oledAddr);
    Wire.write(oldData);
    for (uint8_t x = 0; x < OLED_WIDTH; x++) {
      if (Wire.write(0) == 0) {
        Wire.endTransmission();
        Wire.beginTransmission(oledAddr);
        Wire.write(oldData);
        Wire.write(0);
      }
    }
    Wire.endTransmission();
  }

  oledSendCommand(0xAF) ; // 画面表示
}

// 表示日表示切替
void OLED::dispSwitch(bool sw) {
  if (sw) {
    oledSendCommand(0xAF) ; // 画面表示
  } else {
    oledSendCommand(0xAE) ; // 画面表示
  }
}

// 文字列表示
void OLED::Print(char *str,bool flagCR) {
  int strLen = strlen(str) ;
  uint8_t *strPos = (uint8_t *)str ;
  while (strLen > 0) {
    int utf8len = u2s.UTF8LEN((uint8_t *)strPos) ;
    if (utf8len <= 1) {
      uint8_t asciiCode = *strPos ;
      strPos ++ ;
      strLen -- ;
      DisplayBuff[cursorX + cursorY * 16] = asciiCode ;
      PrintASCII(asciiCode,false);
    } else {
      uint16_t utf16Code = u2s.UTF8ToUTF16(strPos) ;
      uint16_t sjisCode = u2s.UTF16ToSJIS(utf16Code) ;
      strPos += utf8len ;
      strLen -= utf8len ;
      DisplayBuff[cursorX + cursorY * 16] = (sjisCode & 0xFF00) >> 8 ;
      DisplayBuff[cursorX + 1 + cursorY * 16] = sjisCode & 0x00FF ;

      PrintSJIS(sjisCode,false) ;
    }
  }
  if (flagCR) {
    PrintCR() ;
  }
}

// ASCII文字を表示
void OLED::PrintASCII(uint8_t asciiCode,bool flagCR) {
  kanjiRom.readASCIIFont(asciiCode);
  if (cursorX + 1 > OLED_CHAR_W) {
    PrintCR( ) ;
  }
  oledBitMap(cursorX, cursorY, 8, 2, kanjiRom.matrixData) ;
  cursorX ++;
  if (flagCR) {
    PrintCR( ) ;
  }
}

// SJIS漢字を表示
void OLED::PrintSJIS(uint16_t sjisCode,bool flagCR) {
  kanjiRom.readSJISFont(sjisCode);
  if (cursorX + 2 > OLED_CHAR_W) {
    PrintCR( ) ;
  }
  oledBitMap(cursorX, cursorY, 16, 2, kanjiRom.matrixData) ;
  cursorX += 2 ;
  if (flagCR) {
    PrintCR( ) ;
  }
}

// 改行
void OLED::PrintCR( ) {
  if (cursorY >= OLED_CHAR_H - 1) {
    scroll( ) ;  
  } else {
    cursorY +=1 ;
  }
  cursorX = 0 ;
}

// スクロール
void OLED::scroll() {
  memcpy(&DisplayBuff[0],&DisplayBuff[16],16*3) ;
  memset(&DisplayBuff[16*3],0,16) ;

  for (int y = 0 ; y < 4 ; y ++ ) {
    for (int x = 0 ; x < 16 ; x ++) {
      uint8_t code = DisplayBuff[y*16 + x] ;
      if(((code >= 0x81) && (code <= 0x9F)) || ((code >= 0xE0) && (code <= 0xFC))) {
        uint16_t sjisCode = DisplayBuff[y*16 + x] << 8 | DisplayBuff[y*16 + x + 1] ;
        kanjiRom.readSJISFont(sjisCode);
        oledBitMap(x, y, 16, 2, kanjiRom.matrixData) ;
        x++ ;
      } else {
        uint8_t asciiCode = DisplayBuff[y*16 + x] ;
        kanjiRom.readASCIIFont(asciiCode);
        oledBitMap(x, y, 8, 2, kanjiRom.matrixData) ;
      }
    }
  }
}

// OLEDに1バイトコマンドを送信
void OLED::oledSendCommand(uint8_t command) {
  Wire.beginTransmission(oledAddr);
  Wire.write(oldCommand);
  Wire.write(command);
  Wire.endTransmission();
}

// OLEDに2バイトコマンドを送信
void OLED::oledSendCommand2(uint8_t command1,uint8_t command2) {
  Wire.beginTransmission(oledAddr);
  Wire.write(oldCommands);
  Wire.write(command1);
  Wire.write(command2);
  Wire.endTransmission();
}

// OLEDに3バイトコマンドを送信
void OLED::oledSendCommand3(uint8_t command1,uint8_t command2,uint8_t command3) {
  Wire.beginTransmission(oledAddr);
  Wire.write(oldCommands);
  Wire.write(command1);
  Wire.write(command2);
  Wire.write(command3);
  Wire.endTransmission();
}

// OLEDに複数のコマンドを送信::Wireクラスの制限により最大32Byte
// 長さのチェックはしてないのでプログラマ責任で上手い事やってちょ
void OLED::oledSendCommands(int commandLen,uint8_t commands[]) {
  Wire.beginTransmission(oledAddr);
  Wire.write(oldCommands);
  for (int i = 0; i < commandLen; i++) {
    Wire.write(commands[i]);
  }
   Wire.endTransmission();
}

// BITマップを表示する::oledFillDataの可変データ型なので基本oledFillDataと同じ作り
void OLED::oledBitMap(int offsetX , int offsetY , int width, int hight, uint8_t *bitMap) {
  int pos = 0 ;
  for (int m = 0; m < hight; m++) {
    oledBitMap32(offsetX*8 , offsetY * 2 + m , width , &bitMap[pos]) ;    // 上半分
    pos += width ;
  }
  oledSendCommand3(0xB0 , 0x10 , 0x00) ;
}

// 128x32ドットＯＬＥＤにビットマップを書き込み
void OLED::oledBitMap32(int col , int row , int width , uint8_t *bitMap) {
  oledSendCommand3(0xB0 | (row & 0x07) , 0x10 | (col & 0xf0) >> 4 , col & 0x0F) ;
  Wire.beginTransmission(oledAddr);
  Wire.write(oldData);
  for (int x = 0; x < width; x++) {
    if (Wire.write(bitMap[x]) == 0) {
      Wire.endTransmission();
      Wire.beginTransmission(oledAddr);
      Wire.write(oldData);
      Wire.write(bitMap[x]);
    }
  }
  Wire.endTransmission();
}
